#ifndef STK_TX81Z_H
#define STK_TX81Z_H
#include "FM2.h"
namespace stk {
/*! \class TX81Z
    \brief STK TX81Z FM synthesis instrument.
    This class implements algorithms 1 thru 8 of the TX81Z.
    -----------------------------------------------
    Algorithm Structures:
    -----------------------------------------------
    Carrier = first factor of each first level term
    Modulation is by right-multiplication
    FeedBack In  = ]
    FeedBack Out = [
    -----------------------------------------------
    1 - 012[3]
    2 - 01(2+[3])
    3 - 0(12+[3])
    4 - 0(1+2[3])
    5 - 01+2[3]
    6 - 0[3]+1[3]+2[3]
    7 - 0+1+2[3]
    8 - 0+1+2+[3]
    -----------------------------------------------
    Control Change Numbers: 
       - Total Modulator Index = 2
       - Modulator Crossfade = 4
       - LFO Speed = 11
       - LFO Depth = 1
       - ADSR 2 & 4 Target = 128
*/
class TX81Z : public FM2
{
 public:
  TX81Z( void );
  ~TX81Z( void );
  void setFrequency( StkFloat frequency );
  void noteOn( StkFloat frequency, StkFloat amplitude );
  StkFloat tick( unsigned int channel = 0 );  
  StkFrames& tick( StkFrames& frames, unsigned int channel = 0 );
 protected:
};
inline StkFloat TX81Z :: tick( unsigned int )
{
  register StkFloat temp;
  temp = 1.0 + vibrato_.tick() * modDepth_ * 0.1;    
  waves_[0]->setFrequency(baseFrequency_ * temp * ratios_[0]);
  waves_[1]->setFrequency(baseFrequency_ * temp * ratios_[1]);
  waves_[2]->setFrequency(baseFrequency_ * temp * ratios_[2]);
  waves_[3]->setFrequency(baseFrequency_ * temp * ratios_[3]);
  //int algorithm = 5;
  switch(algorithm_) 
  {
      case 1:  //    1 - 012[3]
                         waves_[3]->addPhaseOffset( twozero_.lastOut() );
               temp   =  gains_[3] * adsr_[3]->tick() * waves_[3]->tick();
                         twozero_.tick(temp);
                         waves_[2]->addPhaseOffset( temp );
               temp   =  gains_[2] * adsr_[2]->tick() * waves_[2]->tick();
                         waves_[1]->addPhaseOffset( temp );
               temp   =  gains_[1] * adsr_[1]->tick() * waves_[1]->tick();
                         waves_[0]->addPhaseOffset( temp );
               temp   =  gains_[0] * adsr_[0]->tick() * waves_[0]->tick();
               break;
      case 2:  //    2 - 01(2+[3])
                         waves_[3]->addPhaseOffset( twozero_.lastOut() );
               temp   =  gains_[3] * adsr_[3]->tick() * waves_[3]->tick();
                         twozero_.tick(temp);
               temp  +=  gains_[2] * adsr_[2]->tick() * waves_[2]->tick();
                         waves_[1]->addPhaseOffset( temp );
               temp   =  gains_[1] * adsr_[1]->tick() * waves_[1]->tick();
                         waves_[0]->addPhaseOffset( temp );
               temp   =  gains_[0] * adsr_[0]->tick() * waves_[0]->tick();
               break;
      case 3:  //    3 - 0(12+[3])
               temp   =  gains_[2] * adsr_[2]->tick() * waves_[2]->tick();
                         waves_[1]->addPhaseOffset( temp );
                         waves_[3]->addPhaseOffset( twozero_.lastOut() );
               temp   =  gains_[3] * adsr_[3]->tick() * waves_[3]->tick();
                         twozero_.tick(temp);
               temp  +=  gains_[1] * adsr_[1]->tick() * waves_[1]->tick();
                         waves_[0]->addPhaseOffset( temp );
               temp   =  gains_[0] * adsr_[0]->tick() * waves_[0]->tick();
               break;
      case 4: //    4 - 0(1+2[3])
                         waves_[3]->addPhaseOffset( twozero_.lastOut() );
               temp   =  gains_[3] * adsr_[3]->tick() * waves_[3]->tick();
                         twozero_.tick(temp);
                         waves_[2]->addPhaseOffset( temp );
               temp   =  gains_[2] * adsr_[2]->tick() * waves_[2]->tick();
               temp  +=  gains_[1] * adsr_[1]->tick() * waves_[1]->tick();
                         waves_[0]->addPhaseOffset(temp);
               temp   =  gains_[0] * adsr_[0]->tick() * waves_[0]->tick();
               break;
      case 5:  //    5 - 01+2[3]  
               temp   =  gains_[1] * adsr_[1]->tick() * waves_[1]->tick();
                         waves_[0]->addPhaseOffset( temp );
                         waves_[3]->addPhaseOffset( twozero_.lastOut() );
               temp   =  gains_[3] * adsr_[3]->tick() * waves_[3]->tick();
                         twozero_.tick(temp);
                         waves_[2]->addPhaseOffset( temp );
               temp   =  gains_[0] * adsr_[0]->tick() * waves_[0]->tick();
               temp  +=  gains_[2] * adsr_[2]->tick() * waves_[2]->tick();
               break;
      case 6: //    6 - 0[3]+1[3]+2[3]
                         waves_[3]->addPhaseOffset( twozero_.lastOut() );
               temp   =  gains_[3] * adsr_[3]->tick() * waves_[3]->tick();
                         twozero_.tick( temp );
                         waves_[0]->addPhaseOffset( temp );
                         waves_[1]->addPhaseOffset( temp );
                         waves_[2]->addPhaseOffset( temp );
               temp   =  gains_[0] * adsr_[0]->tick() * waves_[0]->tick();
               temp  +=  gains_[1] * adsr_[1]->tick() * waves_[1]->tick();
               temp  +=  gains_[2] * adsr_[2]->tick() * waves_[2]->tick();
               break;
      case 7:  //    7 - 0+1+2[3]
                         waves_[3]->addPhaseOffset( twozero_.lastOut() );
               temp   =  gains_[3] * adsr_[3]->tick() * waves_[3]->tick();
                         twozero_.tick( temp );
                         waves_[2]->addPhaseOffset( temp );
               temp   =  gains_[0] * adsr_[0]->tick() * waves_[0]->tick();
               temp  +=  gains_[1] * adsr_[1]->tick() * waves_[1]->tick();
               temp  +=  gains_[2] * adsr_[2]->tick() * waves_[2]->tick();
               break;
      case 8:  //    8 - 0+1+2+[3]  
                         waves_[3]->addPhaseOffset( twozero_.lastOut() );
               temp   =  gains_[3] * adsr_[3]->tick() * waves_[3]->tick();
                         twozero_.tick( temp );
               temp  +=  gains_[2] * adsr_[2]->tick() * waves_[2]->tick();
               temp  +=  gains_[1] * adsr_[1]->tick() * waves_[1]->tick();
               temp  +=  gains_[0] * adsr_[0]->tick() * waves_[0]->tick();
               break;
      default:
              break;
  }   
               temp          = temp * (1.0 + vibrato_.tick() * modDepth_);  // amplitude modulation
               lastFrame_[0] = masterVolume_ * temp;
               return lastFrame_[0];
}
inline StkFrames& TX81Z :: tick( StkFrames& frames, unsigned int channel )
{
  unsigned int nChannels = lastFrame_.channels();
  StkFloat *samples = &frames[channel];
  unsigned int j, hop = frames.channels() - nChannels;
  if ( nChannels == 1 ) 
  {
    for ( unsigned int i=0; i<frames.frames(); i++, samples += hop )
      *samples++ = tick();
  }
  else 
  {
    for ( unsigned int i=0; i<frames.frames(); i++, samples += hop ) 
    {
      *samples++ = tick();
      for ( j=1; j<nChannels; j++ )
        *samples++ = lastFrame_[j];
    }
  }
  return frames;
}
} // stk namespace
#endif
